@@ -1,15 +1,15 @@
+import re
+from functools import lru_cache
+import yaml
+import os
+from retry import retry
+import sys
+from rc.machine import Machine
 from rc.util import run
 from rc.exception import MachineCreationException, MachineNotRunningException, MachineShutdownException, \
     MachineDeletionException, MachineChangeTypeException, MachineNotReadyException, SaveImageException, \
-    DeleteImageException, FirewallRuleCreationException, FirewallRuleDeleteionException
-from rc.machine import Machine
-import sys
-from retry import retry
-import os
-import yaml
-from functools import lru_cache
-import re
-import sys
+    DeleteImageException, FirewallRuleCreationException, FirewallRuleDeleteionException, MachineBootupException, \
+    RcException, DiskCreationException, DiskDeletionException, MachineAddDiskException, MachineRemoveDiskException
 
 gcloud_provider = sys.modules[__name__]
 
@@ -52,6 +52,15 @@ def _release_ip_address(name, region):
 
 SSH_KEY_PATH = os.path.expanduser('~/.ssh/google_compute_engine')
 
+if not os.path.exists(SSH_KEY_PATH):
+    print("There is no key at {}, creating new key.".format(SSH_KEY_PATH))
+    # Generate new ssh key to log in.
+    # WARNING: It will create a key with an empty passphrase.
+    run(["ssh-keygen", "-f", SSH_KEY_PATH, "-t", "rsa", "-N", '""'])
+    # Upload key
+    run(["gcloud", "compute", "os-login", "ssh-keys",
+         "add", "--key-file={}.pub".format(SSH_KEY_PATH)])
+
 
 def list():
     p = run(['gcloud', 'compute', 'instances', 'list', '--format',
@@ -97,14 +106,7 @@ def _wait_bootup(name):
         raise MachineNotRunningException(status)
 
 
-@retry(MachineNotReadyException)
-def _wait_ssh(machine):
-    p = machine.run('echo a')
-    if p.returncode != 0:
-        raise MachineNotReadyException(p.stderr)
-
-
-def create(*, name, machine_type, disk_size, image_project, image_family=None, image=None, zone, min_cpu_platform=None,
+def create(name, *, project=None, machine_type, disk_size, image_project, image_family=None, image=None, zone, min_cpu_platform=None,
            preemptible=False, firewall_allows=None, reserve_ip=True, firewalls=None, disk_type=None):
     args = [name]
     args += ['--machine-type', machine_type]
@@ -153,7 +155,7 @@ def create(*, name, machine_type, disk_size, image_project, image_family=None, i
             raise MachineCreationException(p.stderr)
     machine = Machine(provider=gcloud_provider, name=name, zone=zone,
                       ip=ip, username=_get_username(), ssh_key_path=SSH_KEY_PATH)
-    _wait_ssh(machine)
+    machine.wait_ssh()
     return machine
 
 
@@ -187,8 +189,8 @@ def bootup(machine):
     p = run(['gcloud', 'compute', 'instances', 'start',
              machine.name, '--zone', machine.zone])
     if p.returncode != 0:
-        raise MachineShutdownException(p.stderr)
-    _wait_ssh(machine)
+        raise MachineBootupException(p.stderr)
+    machine.wait_ssh()
 
 
 def status(machine):
@@ -221,3 +223,81 @@ def delete_image(image):
     p = run(command)
     if p.returncode != 0:
         raise DeleteImageException(p.stderr)
+
+
+def add_firewall(machine, firewall):
+    pass
+
+
+def remove_firewall(machine, firewall):
+    pass
+
+
+def create_disk(name, *, type='pd-standard', size, project=None, zone):
+    if project:
+        project = f'--project {project}'
+    else:
+        project = ''
+    p = run(
+        f'gcloud beta compute disks create {name} {project} --type={type} --size={size} --zone={zone}')
+    if p.returncode != 0:
+        raise DiskCreationException(p.stderr)
+
+
+def add_disk(machine, disk):
+    p = run(
+        f'gcloud compute instances attach-disk {machine.name} --disk={disk} --zone {machine.zone} --device-name={disk}')
+    if p.returncode != 0:
+        raise MachineAddDiskException(p.stderr)
+    p = machine.sudo('lsblk -f /dev/disk/by-id/google-' +
+                     disk + " | awk '{print $2}'")
+    if p.returncode != 0:
+        raise MachineAddDiskException(p.stderr)
+    fstype = p.stdout.strip()
+    assert fstype.startswith('FSTYPE')
+    if fstype.strip() != 'FSTYPE':
+        # it's already been formatted and there's a filesystem on it
+        p = machine.sudo(f'''
+        mkdir -p /mnt/{disk}
+        mount -o discard,defaults /dev/disk/by-id/google-{disk} /mnt/{disk}
+        chmod a+w /mnt/{disk}
+        ''')
+    else:
+        p = machine.sudo(f'''
+        sudo mkfs.ext4 -m 0 -F -E lazy_itable_init=0,lazy_journal_init=0,discard /dev/disk/by-id/google-{disk}
+        mkdir -p /mnt/{disk}
+        mount -o discard,defaults /dev/disk/by-id/google-{disk} /mnt/{disk}
+        sudo chmod a+w /mnt/{disk}
+        ''')
+    if p.returncode != 0:
+        raise MachineAddDiskException(p.stderr)
+
+
+def remove_disk(machine, disk):
+    p = machine.sudo(f'umount /dev/disk/by-id/google-{disk}')
+    if p.returncode != 0 and 'No such file or directory' not in p.stderr:
+        raise MachineRemoveDiskException(p.stderr)
+    p = run(
+        f'gcloud compute instances detach-disk {machine.name} --disk {disk} --zone {machine.zone}'
+    )
+    if p.returncode != 0:
+        raise MachineRemoveDiskException(p.stderr)
+
+
+def delete_disk(disk, zone):
+    p = run(f'gcloud compute disks delete {disk} --zone {zone}', input='y\n')
+    if p.returncode != 0:
+        raise DiskDeletionException(p.stderr)
+
+
+def set_project(project):
+    p = run(f'gcloud config set project {project}')
+    if p.returncode != 0:
+        raise RcException(p.stderr)
+
+
+def get_project():
+    p = run(f'gcloud config get-value project')
+    if p.returncode != 0:
+        raise RcException(p.stderr)
+    return p.stdout.strip()
